// Copyright (c) 2022, Paurini Taketakehikuroa Wiringi. All rights reserved.
// See `LICENSE`.

import 'dart:async';
import 'widget_states.dart';
import 'package:flutter/material.dart';
import 'package:appwrite/models.dart';
import 'package:appwrite/appwrite.dart' as appwrite;
import 'package:appwrite/src/client_io.dart';

/// A `CollectionWidget` allows for refined handling of an
/// Appwrite collection query and how we display that result
/// in the UI; depending on the state of the query and the
/// result itself.
///
/// It also provides options for pagination, realtime, start &
/// end bumpers, error handling and your usual appwrite query
/// abilities.
///
/// See the [README](https://github.com/p4-k4/appwrite_flutter_widgets#readme) for examples.
class CollectionWidget<T> extends StatefulWidget {
  /// A `CollectionWidget` allows for refined handling of an
  /// Appwrite collection query and how we display that result
  /// in the UI; depending on the state of the query and the
  /// result itself.
  ///
  /// It also provides options for pagination, realtime, start &
  /// end bumpers, error handling and your usual appwrite query
  /// abilities.
  ///
  /// See the [README](https://github.com/p4-k4/appwrite_flutter_widgets#readme) for examples.

  const CollectionWidget(
      {Key? key,
      required this.database,
      required this.collectionId,
      required this.itemsPerPage,
      this.realtime = true,
      this.pageController,
      this.timeout = 10000,
      this.queries,
      this.orderTypes,
      this.orderAttributes,
      required this.onData,
      this.onEmpty,
      this.onWaiting,
      this.onTimeout,
      this.onError,
      this.fallback = const SizedBox.shrink(),
      this.onStartCallback,
      this.onEndCallback})
      : super(key: key);

  /// An instance of `Databases` for the database we're querying.
  final appwrite.Databases database;

  /// The collectionId that we're querying.
  final String collectionId;

  /// Whether or not server-side updates are automatically reflected
  /// on the client/UI.
  final bool realtime;

  /// The maximum number of documents per page.
  final int itemsPerPage;

  /// The `CollectionPageController` that controlls this `CollectionWidget`.
  final CollectionPageController? pageController;

  /// The timeout stopwatch in milliseonds.
  /// A`TimeoutException` is thrown after this time of which at that point,
  /// this `CollectionWidget`'s state will change to `WidgetStates.timeout`.
  final int timeout;

  /// Array of query strings generated using the Query class provided by the
  /// SDK. Learn more about queries. Maximum of 100 queries are allowed, each
  /// 4096 characters long.
  final List<dynamic>? queries;

  /// Array of order directions for sorting attribtues. Possible values are
  /// DESC for descending order, or ASC for ascending order. Maximum of 100
  /// order types are allowed.
  final List<String>? orderTypes;

  /// Array of attributes used to sort results. Maximum of 100 order
  /// attributes are allowed, each 4096 characters long.
  final List<String>? orderAttributes;

  /// A calback that returns a `Widget` to display the data of a query.
  final Widget Function(List<T> data) onData;

  /// A widget to be displayed in the case that the internal document list
  /// is empty (or a collection is empty).
  ///
  /// **Note**: Leaving this as null will create a fallback pattern of:
  /// 1. If `onEmpty` is null, fallback to `onData` if it's not null.
  /// 2. Otherwise, fallback to the `fallbackWidget` if it's not null.
  /// 3. Or as last resort, fallback to a `SizedBox.shrink()` (essentialy blank space).
  /// Therefore, it's encouraged to handle this state althoguh it is optional.
  final Widget? onEmpty;

  /// A widget to be displayed in the case that a request to the appwrite
  /// server is in progress and we're awaiting results.
  ///
  /// **Note**: Leaving this as null will create a fallback pattern of:
  /// 1. If `onWaiting` is null, fallback to `onData` if it's not null.
  /// 2. Otherwise, fallback to the `fallbackWidget` if it's not null.
  /// 3. Or as last resort, fallback to a `SizedBox.shrink()` (essentialy blank space).
  /// Therefore, it's encouraged to handle this state althoguh it is optional.
  final Widget? onWaiting;

  /// A widget to be displayed in the case that a `TimeoutException` occurs.
  ///
  /// **Note**: Leaving this as null will create a fallback pattern of:
  /// 1. If `onWaiting` is null, fallback to `onData` if it's not null.
  /// 2. Otherwise, fallback to the `fallbackWidget` if it's not null.
  /// 3. Or as last resort, fallback to a `SizedBox.shrink()` (essentialy blank space).
  /// Therefore, it's encouraged to handle this state althoguh it is optional.
  final Widget? onTimeout;

  /// A callback that returns a widget to be displayed in the case of an error.
  /// The callback also provides access to the error and stacktrace.
  ///
  /// **Note**: Leaving this as null will create a fallback pattern of:
  /// 1. If `onError` is null, fallback to `onData` if it's not null.
  /// 2. Otherwise, fallback to the `fallbackWidget` if it's not null.
  /// 3. Or as last resort, fallback to a `SizedBox.shrink()` (essentialy blank space).
  /// Therefore, it's encouraged to handle this state althoguh it is optional.
  final Widget Function(Object? error, StackTrace? stackTrace)? onError;

  /// A fallback widget to be displayed if a widget has not been provided where it
  /// is optional.
  final Widget fallback;

  /// A callback triggered when we hit the start of the query.
  final void Function(List<T> data)? onStartCallback;

  /// A callback triggered when we hit the end of the query.
  final void Function(List<T> data)? onEndCallback;

  @override
  CollectionWidgetState<T> createState() => CollectionWidgetState<T>();
}

class CollectionWidgetState<T> extends State<CollectionWidget<T>> {
  DocumentList _documentList = DocumentList(total: 0, documents: <Document>[]);
  int _offset = 0;
  late appwrite.Realtime _realtime;
  late appwrite.RealtimeSubscription _realtimeSubscription;
  PageControllerDirection _pageControllerDirection =
      PageControllerDirection.initial;
  late StreamSubscription<PageControllerDirection> _pageControllerSubscription;
  WidgetStates _widgetStates = WidgetStates.noData;
  Object? _error;
  StackTrace? _stackTrace;

  Future<void> listDocuments() async {
    // print('Offset: ${_offset + widget.itemsPerPage}');
    // print('Documents total: ${_documentList.total}');
    // print('Documents length: ${_documentList.documents.length}');
    if (_pageControllerDirection == PageControllerDirection.previous) {
      // START BUMPER
      _offset = _offset - widget.itemsPerPage;
      if (_offset <= 0) {
        _offset = 0;
        widget.onStartCallback != null
            ? widget.onStartCallback!(_documentList.documents as List<T>)
            : null;
        // Bumper on empty.
        if (_documentList.total == 0) return;
        widget.pageController?._isAtStart = true;
        widget.pageController?._isAtEnd = false;
      }
    } else if (_pageControllerDirection == PageControllerDirection.next) {
      // END BUMPER
      if (_offset + _documentList.documents.length >= _documentList.total) {
        widget.onEndCallback != null
            ? widget.onEndCallback!(_documentList.documents as List<T>)
            : null;
        widget.pageController?._isAtEnd = true;
        widget.pageController?._isAtStart = false;
        return;
      }
      _offset = _offset + widget.itemsPerPage;
      if (_offset >= _documentList.total) _offset = _offset - 1;
    } else if (_pageControllerDirection ==
        PageControllerDirection.updateCurrent) {}

    try {
      _error = null;
      _stackTrace = null;
      setState(() => _widgetStates = WidgetStates.waiting);

      final documentList = await widget.database.listDocuments(
        collectionId: widget.collectionId,
        limit: widget.itemsPerPage,
        offset: _offset,
        orderTypes: widget.orderTypes,
        orderAttributes: widget.orderAttributes,
        queries: widget.queries,
      );
      _documentList = documentList;

      setState(() => _widgetStates = WidgetStates.data);
    } catch (e, s) {
      _error = e;
      _stackTrace = s;
      if (e is TimeoutException) {
        setState(() => _widgetStates = WidgetStates.timeout);
      } else {
        setState(() => _widgetStates = WidgetStates.error);
      }
      rethrow;
    }
  }

  @override
  void didUpdateWidget(covariant CollectionWidget<T> oldWidget) {
    super.didUpdateWidget(oldWidget);
  }

  void init() async {
    // Temprorary workaround for issue:
    // https://github.com/appwrite/sdk-generator/issues/417
    if (!(widget.database.client as ClientIO).initialized) {
      await (widget.database.client as ClientIO).init();
    }

    // Set the initial set of documents and offset.
    final documentList = await widget.database
        .listDocuments(
      collectionId: widget.collectionId,
      limit: widget.itemsPerPage,
      offset: _offset,
      orderTypes: widget.orderTypes,
      orderAttributes: widget.orderAttributes,
      queries: widget.queries,
    )
        .timeout(Duration(milliseconds: widget.timeout), onTimeout: () {
      setState(() => _widgetStates = WidgetStates.timeout);
      return _documentList;
    });
    _documentList = documentList;
    _offset = 0;

    // Bind to the page controller.
    _pageControllerDirection = PageControllerDirection.initial;
    _pageControllerSubscription =
        widget.pageController!.controller.stream.listen((event) async {
      _pageControllerDirection = event;
      await listDocuments().timeout(Duration(milliseconds: widget.timeout),
          onTimeout: () {
        setState(() => _widgetStates = WidgetStates.timeout);
        return _documentList;
      });
    });

    // Bind to the realtime subscription.
    if (widget.realtime) {
      widget.pageController?._isAtStart = true;
      widget.pageController?._isAtEnd = false;

      _realtime = appwrite.Realtime(widget.database.client);
      _realtimeSubscription = _realtime.subscribe([
        'databases.${widget.database.databaseId}.collections.${widget.collectionId}.documents',
        'databases.${widget.database.databaseId}.collections.${widget.collectionId}.documents.*.create',
        'databases.${widget.database.databaseId}.collections.${widget.collectionId}.documents.*.delete',
        'databases.${widget.database.databaseId}.collections.${widget.collectionId}.documents.*.update',
      ]);
      _realtimeSubscription.stream.listen((event) async {
        _pageControllerDirection = PageControllerDirection.updateCurrent;
        await listDocuments().timeout(Duration(milliseconds: widget.timeout),
            onTimeout: () {
          setState(() => _widgetStates = WidgetStates.timeout);
          return _documentList;
        });
      });
    } else {
      null;
    }
    if (_widgetStates == WidgetStates.timeout) {
      setState(() => _widgetStates = WidgetStates.timeout);
    } else if (_documentList.total <= 0) {
      setState(() => _widgetStates = WidgetStates.empty);
      // Start/End are true for empty document lists.
      widget.pageController?._isAtStart = true;
      widget.pageController?._isAtEnd = true;
    } else {
      setState(() => _widgetStates = WidgetStates.data);
    }
  }

  @override
  void initState() {
    WidgetsBinding.instance.addPostFrameCallback((_) => init());
    super.initState();
  }

  @override
  void dispose() {
    _realtimeSubscription.close;
    _pageControllerSubscription.cancel();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    if (_widgetStates == WidgetStates.empty) {
      if (widget.onEmpty != null) {
        return widget.onEmpty!;
      } else if (_documentList.documents.isNotEmpty) {
        return widget.onData(_documentList.documents as List<T>);
      } else {
        return widget.fallback;
      }
    } else if (_widgetStates == WidgetStates.waiting) {
      if (widget.onWaiting != null) {
        return widget.onWaiting!;
      } else if (_documentList.documents.isNotEmpty) {
        return widget.onData(_documentList.documents as List<T>);
      } else {
        return widget.fallback;
      }
    } else if (_widgetStates == WidgetStates.timeout) {
      if (widget.onTimeout != null) {
        return widget.onTimeout!;
      } else if (_documentList.documents.isNotEmpty) {
        return widget.onData(_documentList.documents as List<T>);
      } else {
        return widget.fallback;
      }
    } else if (_widgetStates == WidgetStates.error) {
      if (widget.onError != null) {
        return widget.onError!(_error, _stackTrace);
      } else if (_documentList.documents.isNotEmpty) {
        return widget.onData(_documentList.documents as List<T>);
      } else {
        return widget.fallback;
      }
    } else if (_widgetStates == WidgetStates.data) {
      return widget.onData(_documentList.documents as List<T>);
    } else {
      return widget.fallback;
    }
  }
}

/// Represents the trigger direction of a `CollectionPageController`.
enum PageControllerDirection {
  /// Trigger a `CollectionWidget` to move to the next page.
  next,

  /// Trigger a `CollectionWidget` to move to the previous page.
  previous,

  /// Inidicates a `CollectionWidget` has been initialized and is in an initial state.
  initial,

  /// Triggers the refresh of a `CollectionWidget`'s current page.
  updateCurrent,
}

/// A controller that triggers the positioning and state of a
/// `CollectionWidget` while also holding data related to the
/// current positioning of a `CollectionWidget`.
///
/// A `CollectionPageController` is required for pagination to
/// work with a `CollectionWidget`.
///
/// **Note**: Do not create an instance of this class inside
/// your `Widget.build()` method.
class CollectionPageController {
  /// A controller that triggers the positioning and state of a
  /// `CollectionWidget` while also holding data related to the
  /// current positioning of a `CollectionWidget`.
  ///
  /// A `CollectionPageController` is required for pagination to
  /// work with a `CollectionWidget`.
  ///
  /// **Note**: Do not create an instance of this class inside
  /// your `Widget.build()` method.
  CollectionPageController();

  /// Private - Indicates if we're at the start of the queried documents.
  ///
  /// **Note**: `_isAtStart` and `_isAtEnd` will both be true when the
  /// internal document list of a `CollectionWidget` is empty.
  bool _isAtStart = false;

  /// Private - Indicates if we're at the end of the queried documents.
  ///
  /// **Note**: `_isAtStart` and `_isAtEnd` will both be true when the
  /// internal document list of a `CollectionWidget` is empty.
  bool _isAtEnd = false;

  /// The `StreamController` for this `CollectionPageController`.
  ///
  /// The closing of this `StreamController` is outside of the
  /// responsibility of a `CollectionWidget` therefore, you must
  /// close this controller manually to avoid leaks.
  final controller = StreamController<PageControllerDirection>.broadcast();

  /// Moves offset position backward and gets the previous page of documents.
  get previous => controller.add(PageControllerDirection.previous);

  /// Moves offset position forward and gets the next page of documents.
  get next => controller.add(PageControllerDirection.next);

  /// Refreshes documents for the current page.
  get refresh => controller.add(PageControllerDirection.updateCurrent);

  /// Indicates if we're at the start of the queried documents.
  ///
  /// **Note**: `_isAtStart` and `_isAtEnd` will both be true when the
  /// internal document list of a `CollectionWidget` is empty.
  get isAtStart => _isAtStart;

  /// Indicates if we're at the end of the queried documents.
  ///
  /// **Note**: `_isAtStart` and `_isAtEnd` will both be true when the
  /// internal document list of a `CollectionWidget` is empty.
  get isAtEnd => _isAtEnd;
}
